/*-------------------------------------------------------------------------
 *
 * htup.h
 *	  POSTGRES heap tuple definitions.
 *
 *
 * Portions Copyright (c) 1996-2016, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 *
 * src/include/access/htup.h
 *
 *-------------------------------------------------------------------------
 */
#ifndef HTUP_H
#define HTUP_H

#include "storage/itemptr.h"

/* typedefs and forward declarations for structs defined in htup_details.h */

typedef struct HeapTupleHeaderData HeapTupleHeaderData;

typedef HeapTupleHeaderData *HeapTupleHeader;

typedef struct MinimalTupleData MinimalTupleData;

typedef MinimalTupleData *MinimalTuple;

/* rbt: Проверить существование мандатной метки у записи */
#define HeapTupleHeaderHasMAC(tup) \
	(((tup)->t_infomaskpgac & HEAP_PGAC_HAS_MAC) != 0)

/* rbt: Проверить CCR = false у записи системной таблицы (по умолчанию CCR = true для всех) */
#define HeapTupleHeaderHasMACNCCR(tup) \
( \
	(HeapTupleHeaderHasMAC (tup)) ? \
		(tup)->t_infomaskpgac & HEAP_PGAC_MAC_NCCR \
	: 0 \
)

/* rbt: Установить CCR = false у записи системной таблицы */
#define HeapTupleHeaderSetMACNCCR(tup) \
do { \
	Assert(HeapTupleHeaderHasMAC (tup)); \
	(tup)->t_infomaskpgac |= HEAP_PGAC_MAC_NCCR; \
} while (0)

/* rbt: Сбросить CCR = false у записи системной таблицы */
#define HeapTupleHeaderResetMACNCCR(tup) \
do { \
	Assert(HeapTupleHeaderHasMAC (tup)); \
	(tup)->t_infomaskpgac &= ~HEAP_PGAC_MAC_NCCR; \
} while (0)

/* rbt: Получить указатель на мандатную метку записи */
#define HeapTupleHeaderGetMAC_P(tup) \
( \
	(HeapTupleHeaderHasMAC (tup)) ? \
		(pgmac_label_t*) \
			((char*)(tup) + (tup)->t_hoff - \
				(((tup)->t_infomask & HEAP_HASOID) ? sizeof (Oid) : 0) - sizeof (pgmac_label_t)) \
	: \
		NULL \
)

/* rbt: Проверить признак установленности мандатной метки */
#define HeapTupleHeaderMACIsValid(tup) \
( \
	(HeapTupleHeaderHasMAC (tup)) ? \
		(tup)->t_infomaskpgac & HEAP_PGAC_MAC_VALID \
	: 0 \
)

/* rbt: Установить признак установленности мандатной метки */
#define HeapTupleHeaderSetMACValid(tup) \
do { \
	Assert(HeapTupleHeaderHasMAC (tup)); \
	(tup)->t_infomaskpgac |= HEAP_PGAC_MAC_VALID; \
} while (0)

/* rbt: Сбросить признак установленности мандатной метки */
#define HeapTupleHeaderSetMACInvalid(tup) \
do { \
	Assert(HeapTupleHeaderHasMAC (tup)); \
	(tup)->t_infomaskpgac &= ~HEAP_PGAC_MAC_VALID; \
} while (0)


/* rbt: Установить новое значение мандатной метки записи */
#define HeapTupleHeaderSetMAC_P(tup, mac_p) \
do { \
	HeapTupleHeaderSetMACValid(tup); \
	*HeapTupleHeaderGetMAC_P(tup) = *mac_p; \
} while (0)

/*
 * HeapTupleData is an in-memory data structure that points to a tuple.
 *
 * There are several ways in which this data structure is used:
 *
 * * Pointer to a tuple in a disk buffer: t_data points directly into the
 *	 buffer (which the code had better be holding a pin on, but this is not
 *	 reflected in HeapTupleData itself).
 *
 * * Pointer to nothing: t_data is NULL.  This is used as a failure indication
 *	 in some functions.
 *
 * * Part of a palloc'd tuple: the HeapTupleData itself and the tuple
 *	 form a single palloc'd chunk.  t_data points to the memory location
 *	 immediately following the HeapTupleData struct (at offset HEAPTUPLESIZE).
 *	 This is the output format of heap_form_tuple and related routines.
 *
 * * Separately allocated tuple: t_data points to a palloc'd chunk that
 *	 is not adjacent to the HeapTupleData.  (This case is deprecated since
 *	 it's difficult to tell apart from case #1.  It should be used only in
 *	 limited contexts where the code knows that case #1 will never apply.)
 *
 * * Separately allocated minimal tuple: t_data points MINIMAL_TUPLE_OFFSET
 *	 bytes before the start of a MinimalTuple.  As with the previous case,
 *	 this can't be told apart from case #1 by inspection; code setting up
 *	 or destroying this representation has to know what it's doing.
 *
 * t_len should always be valid, except in the pointer-to-nothing case.
 * t_self and t_tableOid should be valid if the HeapTupleData points to
 * a disk buffer, or if it represents a copy of a tuple on disk.  They
 * should be explicitly set invalid in manufactured tuples.
 */
typedef struct HeapTupleData
{
	uint32		t_len;			/* length of *t_data */
	ItemPointerData t_self;		/* SelfItemPointer */
	Oid			t_tableOid;		/* table the tuple came from */
	HeapTupleHeader t_data;		/* -> tuple header and data */
} HeapTupleData;

typedef HeapTupleData *HeapTuple;

#define HEAPTUPLESIZE	MAXALIGN(sizeof(HeapTupleData))

/*
 * Accessor macros to be used with HeapTuple pointers.
 */
#define HeapTupleIsValid(tuple) PointerIsValid(tuple)

/* HeapTupleHeader functions implemented in utils/time/combocid.c */
extern CommandId HeapTupleHeaderGetCmin(HeapTupleHeader tup);
extern CommandId HeapTupleHeaderGetCmax(HeapTupleHeader tup);
extern void HeapTupleHeaderAdjustCmax(HeapTupleHeader tup,
						  CommandId *cmax, bool *iscombo);

/* Prototype for HeapTupleHeader accessors in heapam.c */
extern TransactionId HeapTupleGetUpdateXid(HeapTupleHeader tuple);

/* rbt: Проверить существование мандатной метки у записи */
#define HeapTupleHasMAC(tup) \
		HeapTupleHeaderHasMAC ((tup)->t_data)

/* rbt: Проверить MAC CCR = false у записи системной таблицы */
#define HeapTupleHasMACNCCR(tup) \
		HeapTupleHeaderHasMACNCCR ((tup)->t_data)

/* rbt: Получить указатель на мандатную метку записи */
#define HeapTupleGetMAC_P(tup) \
		HeapTupleHeaderGetMAC_P((tup)->t_data)

/* rbt: Установить новое значение мандатной метки записи */
#define HeapTupleSetMAC_P(tup, mac) \
		HeapTupleHeaderSetMAC_P((tup)->t_data, (mac))

/* rbt: Проверить признак установленности мандатной метки */
#define HeapTupleMACIsValid(tup) \
		HeapTupleHeaderMACIsValid((tup)->t_data)

/* rbt: Установить признак установленности мандатной метки */
#define HeapTupleSetMACValid(tup) \
		HeapTupleHeaderSetMACValid((tup)->t_data)

/* rbt: Сбросить признак установленности мандатной метки */
#define HeapTupleSetMACInvalid(tup) \
		HeapTupleHeaderSetMACInvalid((tup)->t_data)

/* rbt: Установить CCR = false у записи системной таблицы */
#define HeapTupleSetMACNCCR(tup) \
		HeapTupleHeaderSetMACNCCR((tup)->t_data)

/* rbt: Сбросить CCR = false у записи системной таблицы */
#define HeapTupleResetMACNCCR(tup) \
		HeapTupleHeaderResetMACNCCR((tup)->t_data)

#endif   /* HTUP_H */
