/*-------------------------------------------------------------------------
 *
 * block.h
 *	  POSTGRES disk block definitions.
 *
 *
 * Portions Copyright (c) 1996-2016, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 *
 * src/include/storage/block.h
 *
 *-------------------------------------------------------------------------
 */
#ifndef BLOCK_H
#define BLOCK_H

/**
 * @defgroup block Номер (идентификатор) блока на диске
 * @ingroup storage
 *
 * Все данные в СУБД (таблицы и индексы) хранятся в виде именованной (RelFileNode)
 * последовательности блоков фиксированного размера. Блоки являются единицами
 * ввода/вывода информации, т.к. каждый буфер (Buffer) в оперативной памяти
 * обеспечивает доступ к одному блоку на диске. С другой стороны один блок
 * обеспечивает хранение одной страницы (Page), таким образом обеспечивая
 * хранение собственно данных.
 *
 * Единственными частями кода, которые напрямую используют блочное представление
 * данных являются методы доступа (access methods), менеджер буферов (buffer manager)
 * и менеджер хранилищ (storage manager).
 */
/** @{*/
/*
 * BlockNumber:
 *
 * each data file (heap or index) is divided into postgres disk blocks
 * (which may be thought of as the unit of i/o -- a postgres buffer
 * contains exactly one disk block).  the blocks are numbered
 * sequentially, 0 to 0xFFFFFFFE.
 *
 * InvalidBlockNumber is the same thing as P_NEW in buf.h.
 *
 * the access methods, the buffer manager and the storage manager are
 * more or less the only pieces of code that should be accessing disk
 * blocks directly.
 */
//! Номер блока в таблице
/**
 * Каждый блок в последовательности имеет уникальный номер. Они нумеруются последовательно
 * с @c 0 до @c 0xFFFFFFFE.
 */
typedef uint32 BlockNumber;

/**
 * @name Константы номеров блоков
 */
/** @{*/
//! Неопределенный номер блока
#define InvalidBlockNumber		((BlockNumber) 0xFFFFFFFF)

//! Максимальный номер блока
#define MaxBlockNumber			((BlockNumber) 0xFFFFFFFE)
/** @}*/

/*
 * BlockId:
 *
 * this is a storage type for BlockNumber.  in other words, this type
 * is used for on-disk structures (e.g., in HeapTupleData) whereas
 * BlockNumber is the type on which calculations are performed (e.g.,
 * in access method code).
 *
 * there doesn't appear to be any reason to have separate types except
 * for the fact that BlockIds can be SHORTALIGN'd (and therefore any
 * structures that contains them, such as ItemPointerData, can also be
 * SHORTALIGN'd).  this is an important consideration for reducing the
 * space requirements of the line pointer (ItemIdData) array on each
 * page and the header of each heap or index tuple, so it doesn't seem
 * wise to change this without good reason.
 */
//! Идентификатор блока
/**
 * Этот тип является эквивалентом BlockNumber. Он используется при хранении
 * ссылок на блоки в структурах, которые должны храниться на диске (например
 * HeapTupleHeaderData), тогда как BlockNumber используется при вычислениях
 * (например в методах доступа).
 *
 * Необходимость в дополнительном типе для одной и той же сущности обусловлена
 * тем, что BlockIdData может быть размещен с максимальной плотностью. Это
 * важно для уменьшения объема пространства, занимаемого такими данными, как
 * ItemPointerData.
 */
typedef struct BlockIdData
{
	//! Старшие два байта идентификатора блока
	uint16		bi_hi;
	//! Младшие два байта идентификатора блока
	uint16		bi_lo;
} BlockIdData;

//! Псевдоним для указателя на идентификатор блока
typedef BlockIdData *BlockId;	/* block identifier */

/* ----------------
 *		support macros
 * ----------------
 */

/**
 * @name Макросы для обработки номера (идентификатора) блока
 */
/*
 * BlockNumberIsValid
 *		True iff blockNumber is valid.
 */
//! Проверка номера блока на достоверность
#define BlockNumberIsValid(blockNumber) \
	((bool) ((BlockNumber) (blockNumber) != InvalidBlockNumber))

/*
 * BlockIdIsValid
 *		True iff the block identifier is valid.
 */
//! Проверка идентификатора блока на достоверность
#define BlockIdIsValid(blockId) \
	((bool) PointerIsValid(blockId))

/*
 * BlockIdSet
 *		Sets a block identifier to the specified value.
 */
//! Установить идентификатор блока по номеру блока
#define BlockIdSet(blockId, blockNumber) \
( \
	AssertMacro(PointerIsValid(blockId)), \
	(blockId)->bi_hi = (blockNumber) >> 16, \
	(blockId)->bi_lo = (blockNumber) & 0xffff \
)

/*
 * BlockIdCopy
 *		Copy a block identifier.
 */
//! Скопировать идентификатор блока
#define BlockIdCopy(toBlockId, fromBlockId) \
( \
	AssertMacro(PointerIsValid(toBlockId)), \
	AssertMacro(PointerIsValid(fromBlockId)), \
	(toBlockId)->bi_hi = (fromBlockId)->bi_hi, \
	(toBlockId)->bi_lo = (fromBlockId)->bi_lo \
)

/*
 * BlockIdEquals
 *		Check for block number equality.
 */
//! Проверить идентификаторы блоков на равенство
#define BlockIdEquals(blockId1, blockId2) \
	((blockId1)->bi_hi == (blockId2)->bi_hi && \
	 (blockId1)->bi_lo == (blockId2)->bi_lo)

/*
 * BlockIdGetBlockNumber
 *		Retrieve the block number from a block identifier.
 */
//! Преобразовать идентификатор блока в номер блока
#define BlockIdGetBlockNumber(blockId) \
( \
	AssertMacro(BlockIdIsValid(blockId)), \
	(BlockNumber) (((blockId)->bi_hi << 16) | ((uint16) (blockId)->bi_lo)) \
)

/** @}*/
#endif   /* BLOCK_H */
