#ifndef PGAC_MAC_H
#define PGAC_MAC_H
/**
 * @file
 * @brief Спецификация подсистемы расширенного управления доступом к данным
 * @ingroup iutils
 *
 * В этом модуле собраны объекты, обеспечивающие расширенное управление доступом
 * (в частности реализацию мандатного контроля доступа к данным), которые используются
 * затем в основном коде СУБД.
 *
 * Все эти объекты вынесены в отдельный модуль с тем, чтобы минимизировать правки в
 * остальном коде и соответственно упростить сопровождение кода и переход на новые
 * версии СУБД @c PostgreSQL.
 */

#include "utils/pgac_types.h"
#include "utils/pgac_audit.h"

#include "utils/rel.h"
#include "utils/guc.h"
#include <catalog/objectaddress.h>

#include "storage/large_object.h"

#define __LINUX_POSIX_ACL_H

/**
 * @name Конфигурационные переменные модуля
 */
/** @{*/
//! Флаг, управляющий игнорированием метки входящего соединения
/**
 * Этот флаг обеспечивает возможности работы с секретной информацией из сессий
 * пользователей с произвольными метками (в частности с клиентов Windows всех версий,
 * для которых метка сессии всегда равна {0, 0}).
 *
 * @warning Поскольку после получения секретной информации из базы на такого клиента
 *          ее метка будет утеряна и дальнейшее распространение информации не может
 *          быть проконтролировано, то этот флаг открывает неконтролируемый канал
 *          утечки секретной информации и может использоваться только на этапе
 *          разработки АС. В режиме эксплуатации СУБД его следует установить в
 *          @c false;
 *
 * @todo Определиться со значением по умолчанию (пока @c true).
 */
extern bool pgac_ignore_socket_maclabel;

//! Флаг, управляющий игнорированием метки серверного процесса
/**
 * Этот флаг контролирует защиту данных на стороне сервера и позволяет ограничить сверху
 * уровень хранимой в базе информации (для этого он должен быть установлен в @c false).
 *
 * Если этот параметр установлен в @c false, то при соединении с базой метка процесса
 * сервера (@c postmaster и @c postgres) будет использоваться при определении доступности
 * базы для записи (т.е. для определения прав пользователя на выполнение команд @c INSERT,
 * @c UPDATE и @c DELETE).
 *
 * Кроме этого, метка сервера будет использоваться в процессе авторизации пользователя и
 * если метка пользователя оказалась несовместимой с меткой сервера, то
 * процесс авторизации будет прерван с выдачей сообщения о несовместимости меток и
 * невозможности продолжать работу с базой.
 *
 * Для того, чтобы этот механизм начал функционировать @em ДО инициализации кластера баз
 * данных пользователю @c postgres должна быть установлена соответствующая метка. После
 * этого, в процессе авторизации, все каталоги и файлы кластера (<tt>/var/lib/pgsql/data</tt>)
 * будут созданы с меткой пользователя @c postgres. Если в последствии метка пользователя
 * @c postgres будет изменена, то это приведет к тому, что созданные каталоги будут
 * недоступны, и сервер не будет запускаться (поскольку не сможет получить доступа к своим
 * данным). Это обеспечивает дополительную защиту данных со стороны ОС.
 *
 * Для того, чтобы определение метки сервера и защита данных были однозначными, при
 * задании меток пользователю @c postgres обе метки диапазона должны быть равны (иначе
 * запуска сервера и инициализации каталогов и файлов кластера не произойдет).
 *
 * Если этот флаг установлен в @c true, то метка сервера игнорируется и в базе данных
 * может сохраняться информация с произвольными метками. Метка пользователя @c postgres
 * в этом случае определяет только степень защиты каталогов и файлов кластера баз данных и
 * не влияет на процедуру авторизации пользователей и их работу.
 *
 * @warning Поскольку защита данных со стороны ОС является необходимым условием эксплуатации
 *          СУБД при работе с секретной информацией, то установка этого флага в @c true допустима
 *          только в процессе разработки АС на базе СУБД. В режиме эксплуатации он должен быть
 *          установлен в @c false.
 *
 * @warning Косвенным следствием установки этого флага в @c false является тот факт,
 *          что при установке в @c false флага pgac_ignore_socket_maclabel подключится под
 *          именем суперпользователя @c postgres смогут только пользователи с более
 *          высокой чем у него меткой.
 *
 */
extern bool pgac_ignore_server_maclabel;

//! Флаг, разрешающий владельцам объектов модифицировать права доступа к своим объектам
/**
 * При установке этого флага в @c false пользователи СУБД, не являющиеся администраторами
 * не могут произвольно раздавать права на доступ к объектам, которыми они владеют. (по
 * умолчанию объекты создаются с правом доступа к ним только для своих владельцев). Это
 * позволяет предотвратить распространение прав на доступ к информации (когда пользователь,
 * имеющий доступ на чтение к некоторой информации может предоставить доступ к ней пользователю,
 * не имеющему прав на ознакомление с ней, путем создания промежуточного объекта и предоставления
 * доступа к нему другим пользователям).
 *
 */
extern bool pgac_enable_trusted_owner;

//! Флаг, разрешающий использовать команду @c TRUNCATE для быстрой очистки таблиц
/**
 * С помощью команды @c TRUNCATE владелец таблицы может удалить из нее данные с любыми
 * мандатными метками. Чтобы предотвратить такую ситуацию и введен этот флаг. Установка
 * его в @c false блокирует выполнение этой команды для любых пользователей. Поскольку
 * команда @c TRUNCATE является нестандартным расширением @c PostgreSQL, то она не
 * должна оказывать влияние на приложения, разработанные в рамках стандарта SQL.
 *
 * @warning При эксплуатации СУБД нужно установить этот флаг в значение @c false.
 *
 */
extern bool pgac_enable_truncate;

//! Флаг, управляющий политикой мандатного разграничения доступа к последовательностям (счетчикам)
/**
 * При установке этого флага в @c true разрешается использование признака MAC CCR для последовательностей
 * аналогично таблицам и видам. В противном случае (установка флага в @c false) признак MAC CCR считается
 * всегда установленным.
 * Последовательности могут представлять скрытый канал долступа к информации. В случае, когда
 * это не представляет опасности, возможно отключение использования признака MAC CCR для последовательностей.
 *
 */
extern bool pgac_enable_sequence_ccr;

//! Флаг, разрешающий использовать команду COPY для вывода в файл.
/**
 * При установке этого флага в @c true суперпользователь может выполнять команду @c COPY
 * для вывода информации в файл. Взаимодействие с ОС осуществляется от имени пользователя
 * postgres. Установка его в @c false блокирует выполнение этой команды для любых пользователей.
 * Поскольку команда @c COPY является нестандартным расширением @c PostgreSQL, то она не
 * должна оказывать влияние на приложения, разработанные в рамках стандарта SQL.
 *
 * @warning При эксплуатации СУБД нужно установить этот флаг в значение @c false.
 *
 */
extern bool pgac_enable_copy_to_file;

//! Флаг, разрешающий использовать расширение dblink в условиях мандатного разграничения доступа
/**
 * При установке этого флага в @c true разрешается использовать расширение dblink
 * для доступа к данным с уровнем отличным от нуля
 *
 */
extern bool pgac_enable_dblink_mac;

//! Флаг, разрешающий передавать привилегии пользователям
/**
 * При установке этого флага в @c true разрешается передавать привилегии другим пользователям
 * с помощью команды GRANT
 */
extern bool pgac_enable_grant_options;

//! Флаг, разрешающий передавать право администрирования другим пользователям
/**
 * При установке этого флага в @c true разрешается передавать право администрирования другим пользователям
 * с помощью команды GRANT
 */
extern bool pgac_enable_admin_options;

//! Флаг, разрешающий автоматическую коррекцию метки контейнера.
/**
 * При установке этого флага в @c true в процессе выполнения загрузки данных,
 * например из резервных копий, не содержащих меток контейнеров, метки
 * контейнеров будут подниматься соответственно данным.
 *
 * @warning При эксплуатации СУБД нужно установить этот флаг в значение @c false.
 *
 */
extern bool pgac_auto_adjust_macs;

//! Время жизни (TTL) информации о привилегиях пользователя
/**
 * Полученные привилегии пользователя с целью предотвращения падения производительности
 * кешируются.
 *
 * Параметр задает время жизни информации о привилегиях пользователя, после которого
 * информация о привилегиях пользователя заново запрашивается.
 *
 * Значение равное нулю отключает кеширование.
 *
 */
extern int pgac_caps_ttl;

//! Флаг, управляющий выводом отладочной информации в журнал сервера
/**
 * При установке этого флага в @c on в журнал сервера выводится отладочная информация
 * модуля с соответствующими уровнями ошибок DEBUG1-DEBUG3
 *
 */
extern bool pgac_debug_print;

/** @}*/

/**
 * @name Управление сессией пользователя
 */
/** @{*/

//! Авторизация и инициализация мандатных параметров сессии пользователя
/**
 * @param[in] _username текущее имя пользователя
 * @param[in] _is_suser @b true если пользователь является суперпользователем
 */
void pgac_sessionInit(const char *_username, bool _is_suser);

//! Завершение работы модуля
void pgac_sessionDone(void);

//! Восстановление мандатных атрибутов сессии при завершении транзакции
/**
 * Сброс мандатных параметров сессии в начальное состояние.
 *
 * @param[in] _is_commit @b true - при вызове для подтверждения изменений,
 *                       @b false - для отмены
 */
void pgac_AtEOXactRestore(bool _is_commit);

//! Сохранение мандатных атрибутов сессии перед правкой метаданных
void pgac_preAlterSet(const pgmac_label_t *label, bool macccr);


//! Восстановление мандатных атрибутов сессии при завершении правки метаданных
/**
 *
 */
void pgac_postAlterRestore(void);

/**
 * @name Выполнение блока команд с временной сменой вектора мандатных атрибутов
 *
 * Это необходимо для обеспечения целостности обрабатываемых данных для внутренних
 * нужд сервера. Например при построении индексов, вызове встроенных функций и
 * выполнении вспомогательных работ (например @c VACUUM).
 *
 * Эти функции должны выполняться в режиме полного доступа к данным, так как иначе
 * возможно полное нарушение внутренней логики работы сервера.
 *
 * С точки зрения безопасности это допустимо, если такие функции не модифицируют
 * логическую сущность данных и не взаимодействуют с пользователем путем передачи
 * ему обрабатываемой информации. Выполнение этих условий необходимо контролировать
 * в каждом отдельном случае использования этого механизма.
 *
 * В общем случае применение этого механизма выглядит так:
 *
 * @code
 * {
 *         pgac_SUDO_STORE;
 *         pgac_SUDO(pgmac_CAPS_ADD(pgac_getCaps(), pgmac_CAP_OVERRIDE));
 *
 *         ... блок команд, выполняемых в условиях отмены мандатного контроля
 *
 *         pgac_SUDO_RESTORE;
 * }
 * @endcode
 */
/** @{*/
//! Получить исходный вектор мандатных атрибутов сессии
pgmac_caps_t pgac_getSessionCaps(void);
//! Установить новое значение текущего вектора мандатных атрибутов
void         pgac_setCaps(pgmac_caps_t caps);
//! Получить значение текущего вектора мандатных атрибутов
pgmac_caps_t pgac_getCaps(void);
//! Сохранить текущее значение вектора мандатных атрибутов на стеке
#define pgac_SUDO_STORE   pgmac_caps_t _pg_caps = pgac_getCaps()
//! Установить новое значение вектора мандатных атрибутов
#define pgac_SUDO(a)      pgac_setCaps(a)
//! Восстановить ранее сохраненное на стеке значение вектора мандатных атрибутов
#define pgac_SUDO_RESTORE pgac_setCaps(_pg_caps)
/** @}*/

/**
 * @name Вспомогательные функции для обеспечения мандатного контроля
 */
//! Классификация объектов, создаваемых в системе, на те, записи которых имеют мандатные
//! метки и прочие
/**
 * Эта процедура вызывается в процессе создания таблиц, для решения вопроса об
 * использовании или неиспользовании для нее мандатной защиты.
 *
 * Правила классификации:
 *   -# защищаются строки только объектов, хранящие данные (RELKIND_RELATION);
 *   -# таблицы во временных пространствах имен не защищаем;
 *   -# защита остальных таблиц определяется параметром @a _hint.
 *
 * @param name имя создаваемой таблицы
 * @param namespaceId уникальный идентификатор пространтства имен, в котором
 *                   создается таблица
 * @param relkind тип создаваемой таблицы (см. include/catalog/pg_class.h определение таблицы
 *                 таблиц)
 * @param hint обеспечивает принудительное выключение MAC контроля для заданной таблицы
 *              (в настоящее время не используется, предполагалось, что синтаксис команды
 *               <tt>CREATE TABLE</tt> будет позволять по желанию защищать отдельные таблицы
 *               мандатным контролем, так как это сделано для @c Oid)
 * @return признак необходимости защиты записей таблицы (@c true) или ее отсутствия (@c false)
 */
bool pgac_protectRelationTuples(const char *name, Oid namespaceId, char relkind, bool hint);

//! Инициализация метки бинарного потока (LargeObject) по его дескриптору
/**
 * LargeObject хранятся в таблице @c pg_largeobject в виде последовательности страниц.
 * Атрибуты LargeObject хранятся в таблице @c pg_largeobject_metadata
 *
 * @param obj_desc дескриптор бинарного потока
 */
void pgac_LargeObjectLabel(LargeObjectDesc *obj_desc);

//! Проверка возможности создания бинарного потока
/**
 * @param lobjId   идентификатор нового бинарного потока
 */
void pgac_checkCreateLargeObject(Oid lobjId);

void pgac_checkLargeObjectTuple(Oid lobjId, pgmac_label_t *label);

//! Освобождение ресурсов после работы с бинарными потоками
/**
 * @param isCommit  статус завершения
 */
void pgac_closeLargeObjects(bool isCommit);

//! Проверка доступа для заданной метки и текущего значения вектора мандатных атрибутов
//! для операции чтения
/**
 * @param label проверяемая метка
 */
bool pgac_canRead(const pgmac_label_t *label);

//! Проверка доступа для заданной метки и заданного значения вектора мандатных атрибутов
//! для операции чтения
/**
 * @param label проверяемая метка
 * @param caps  вектор мандатных атрибутов
 */
bool pgac_canReadWithCaps(const pgmac_label_t *label, pgmac_caps_t caps);

//! Проверка доступа для заданной метки и текущего значения вектора мандатных атрибутов
//! для операции записи
/**
 * @param label проверяемая метка
 */
bool pgac_canWrite(const pgmac_label_t *label);

//! Проверка доступа для заданной метки и заданного значения вектора мандатных атрибутов
//! для операции записи
/**
 * @param label проверяемая метка
 * @param caps  вектор мандатных атрибутов
 */
bool pgac_canWriteWithCaps(const pgmac_label_t *label, pgmac_caps_t caps);

//! Установка значения метки по текущей метке сессии
/**
 * @param label модифицируемая метка
 */
void pgac_setLabel(pgmac_label_t *label);

//! Проверка мандатной метки на достоверность при явном задании метки записи
/**
 * Метка должна быть безусловно ниже или равна метке отношения, если оно задано,
 * или метке сервера в противном случае и находиться
 * в пределах диапазона меток пользователя. Проверка на нахождение в пределах
 * диапазона выполняется с учетом текущего значения вектора мандатных атрибутов.
 *
 * @param label проверяемая метка
 */
bool pgac_labelIsValid(const Relation relation, const pgmac_label_t *label);

//! Определение доступности базы для "записи"
/**
 * @return @c true если текущая метка сессии пользователя выше метки сервера
 *         и @c false в остальных случаях
 *
 * Эта процедура определяет, должны ли при текущих мандатных атрибутах пользователя
 * блокироваться операции изменения базы данных INSERT, DELETE, UPDATE и COPY FROM в
 * том случае, если включен контроль метки сервера.
 *
 * @note В текущей версии блокировка внесения изменений совмещена с проверкой флага
 *       XactReadOnly. Это представляется более правильным, нежели блокировка на уровне
 *       прав доступа к таблицам.
 *
 * @note Специально отслеживать флаг учета метки сервера не нужно из-за соответствующей
 *       инициализации метки сервера по-умолчанию (т.е. при отсутствии контроля метка
 *       сервера устанавливается в неопределенное состояние pgmac_DECLARE_UNDEFINED_LABEL,
 *       которое заведомо выше любой возможной метки).
 *
 * @warning При переходе на новые версии проверять появление новых мест проверки флага
 *          XactReadOnly.
 *
 * @todo Поверить работу новой реализации.
 */
bool pgac_databaseIsReadOnly(void);

//! Получение вектора мандатных атрибутов пользователя по его id
/**
 * Эта функция необходима для реализации вызова функций со сменой вектора мандатных
 * атрибутов в <tt>backend/utils/fmgr/fmgr.c</tt> и поверки мандатных прав доступа
 * к таблицам и видам в <tt>backend/catalog/aclchk.c</tt>.
 *
 * @warning Пользователь, зарегистрированный как суперпользователь в базе данных
 *          является также и суперпользователем с точки зрения мандатного контроля
 *          (но только в рамках базы данных, поэтому нет необходимости назначать
 *          ему эти привилегии на уровне ОС). Для этого, в процессе авторизации, после
 *          определения статуса пользователя как суперпользователя, ему устанавливаются
 *          флаги pgmac_CAP_OVERRIDE и флаг CAP_CHMAC, которые позволяют ему выполнять
 *          задачи архивирования и восстановления базы данных из архива. Правила
 *          назначения меток сессии для такого пользователя не меняются и они
 *          устанавливаются для него на общих основаниях.
 *
 * @warning Отличия владельца базы данных от обычного пользователя при соединении со своей
 *          базой данных следующие:
 *            - разрешено комментировать объекты своей базы данных;
 *            - разрешено реиндексировать свою базу данных;
 *            - разрешено выполнять команду @c ANALYZE для всех объектов своей базы данных;
 *            - разрешено выполнять команду @c VACUUM для всех объектов своей базы данных;
 *            .
 *          таким образом владелец базы данных ближе к обычному пользователю, чем к
 *          суперпользователю и дополнительных прав, с точки зрения мандатного контроля,
 *          ему не дается.
 *
 * @todo Концепция суперпользователей СУБД нуждается в дальнейшем уточнении и обосновании.
 *
 * @todo Пока реализация без кэширования, "в лоб". Прежде чем бросаться реализовывать
 *       кэширование посмотреть, не кэшируются-ли FmgrInfo, для заполнения которых
 *       эта функция используется.
 */
pgmac_caps_t pgac_getUserCapsByID(Oid user_id);
/** @}*/

//! Установка нового значения мандатной метки
/**
 * @param stmt   запрос
 */
void ExecMACLabelObject(MACLabelStmt *stmt);

//! Установка нового значения признака MAC CCR
/**
 * @param stmt   запрос
 */
void ExecMACCCRObject(MACCCRStmt *stmt);

//! Проверка доступа к изменению отношения (метаданные)
/**
 * @param reloid   	идентификатор изменяемого объекта pg_class
 * @param event	 	событие доступа
 */
void pgac_checkAlterTable(Oid reloid, pgac_Event event);

//! Проверка доступа к изменению объекта
/**
  * @param classId		класс объекта
  * @param objectId		идентификатор объекта
  * @param event	 	событие доступа
 */
void pgac_checkAlterObject(Oid classId, Oid objectId);

//! Проверка доступа к изменению объекта
/**
  * @param tuple		объект
  * @param event	 	событие доступа
 */
void pgac_checkAlter(const HeapTuple tuple);

//! Проверка доступа к отношению
/**
 * @param reloid		идентификатор объекта pg_class
 * @param roleid		идентификатор субъекта доступа
 * @param mask	 		маска доступа
 * @param checkcolmacs 	необходимость проверки меток столбцов
 */
bool pgac_class_maccheck(Oid reloid, Oid roleid, AclMode mask,
						 bool ereport_on_violation);

//! Проверка доступа к базе данных
/**
 * @param dboid   		идентификатор объекта pg_database
 * @param roleid		идентификатор субъекта доступа
 * @param mask	 		маска доступа
*/
bool pgac_database_maccheck(Oid dboid, Oid roleid, AclMode mask,
							bool ereport_on_violation);
//! Проверка доступа к схеме
/**
 * @param nspoid   		идентификатор объекта pg_namespace
 * @param roleid		идентификатор субъекта доступа
 * @param mask	 		маска доступа
*/
bool pgac_namespace_maccheck(Oid nspoid, Oid roleid, AclMode mask,
							bool ereport_on_violation);

//! Проверка доступа к табличному пространству
/**
 * @param spcoid   		идентификатор объекта pg_tablespace
 * @param roleid		идентификатор субъекта доступа
 * @param mask	 		маска доступа
*/
bool pgac_tablespace_maccheck(Oid spcoid, Oid roleid, AclMode mask,
							bool ereport_on_violation);

//! Получение мандатной метки объекта
/**
 * @param classId		класс объекта
 * @param objectId		идентификатор объекта
 * @param label   		получаемая метка
 */
bool pgac_object_maclabel(Oid classId, Oid objectId, pgmac_label_t *label, bool *macccr);

//! Получение мандатной метки базы данных
/**
 * @param dboid   		идентификатор объекта pg_database
 * @param label   		получаемая метка
 */
void pgac_database_maclabel(Oid dboid, pgmac_label_t * label);

//! Получение мандатной метки базы схемы
/**
 * @param nspoid   		идентификатор объекта pg_namespace
 * @param label   		получаемая метка
 */
void pgac_namespace_maclabel(Oid nspoid, pgmac_label_t * label);

//! Получение мандатной метки табличного пространства
/**
 * @param spcoid   		идентификатор объекта pg_tablespace
 * @param label   		получаемая метка
 */
void pgac_tablespace_maclabel(Oid spcoid, pgmac_label_t * label);

//! Получить метку сервера
void pgac_server_maclabel(pgmac_label_t * label);

//! Установка метки при вставке данных
/**
 * @param relation   отношение
 * @param newtuple   новая версия записи
 */
void pgac_TupleInsert(const Relation relation, HeapTuple newtup);

//! Установка метки при обновлении данных
/**
 * @param relation   отношение
 * @param oldtuple   старая версия записи
 * @param newtuple   новая версия записи
 */
void pgac_TupleUpdate(const Relation relation, HeapTuple oldtup, HeapTuple newtup);

//! Автоматическая коррекция метки контейнеров при восстановлении данных
/**
 * @param relation   отношение
 * @param label   мандатная метка
 */
void pgac_AdjustMacs(const Relation relation, pgmac_label_t * label);

/**
 * @name Интерактивный доступ пользователя к текущим параметрам модуля
 *
 * Интерактивный доступ осуществляется с помощью системны конфигурационных параметров
 * сервера (<tt>guc.c</tt>) с помощью команд <tt>SHOW</tt>/<tt>SET</tt> и представления
 * <tt>pg_settings</tt>.
 */
/** @{*/
//! Текстовое представление минимально возможной метки сессии (доступ только для чтения)
extern char *pgac_user_min_maclabel_str;
//! Текстовое представление максимально возможной метки сессии (доступ только для чтения)
extern char *pgac_user_max_maclabel_str;

//! Текущая метка сессии
extern char *pgac_session_maclabel_str;
//! Проверка возможности установки метки сессии
bool pgac_check_sessionLabel(char **newval, void **extra, GucSource source);
//! Установка текущей метки сессии
void pgac_assign_sessionLabel(const char *newval, void *extra);
//! Отображение текущей метки сессии
const char *pgac_show_sessionLabel(void);

//! "Заглушка" для использования совместно с pgac_show_capable_* функциями
extern bool pgac_show_capable_dummy;

//! Может ли пользователь игнорировать мандатный котроль по уровню конфиденциальности
const char *pgac_show_capable_ignmaclvl(void);

//! Может ли пользователь игнорировать мандатный котроль по вектору категорий доступа
const char *pgac_show_capable_ignmaccat(void);

//! Может ли пользователь игнорировать мандатный котроль при операции чтения
const char *pgac_show_capable_mac_readsearch(void);

//! Может ли пользователь менять текущую метку сессии
const char *pgac_show_capable_setmac(void);

//! Может ли пользователь менять метки данных
const char *pgac_show_capable_chmac(void);
/** @}*/

//! Объект является кластером?
bool pgac_is_cluster(const Oid objectid);

/**
 * @name Функции доступные для пользователя из SQL
 *
 */
/** @{*/
//! Получение текущей метки сессии SESSION_MACLABEL
/**
 * @return указатель на мандатную метку (prfee)
 */
Datum pgmac_session_maclabel(PG_FUNCTION_ARGS);
/** @}*/

/** @{*/
//! Получение привилегий сессии SESSION_CAPS
/**
 * @return int64
 */
Datum pgmac_session_caps(PG_FUNCTION_ARGS);
/** @}*/

/** @{*/
//! Получение текущих привилегий CURRENT_CAPS
/**
 * @return int64
 */
Datum pgmac_current_caps(PG_FUNCTION_ARGS);
/** @}*/

/** @{*/
//! Получение метки кластера CLUSTER_MACLABEL
/**
 * @return указатель на мандатную метку (prfee)
 */
Datum pgmac_cluster_maclabel(PG_FUNCTION_ARGS);
/** @}*/

/** @{*/
//! Получение флага CCR кластера CLUSTER_MACCCR
/**
 * @return boolean
 */
Datum pgmac_cluster_macccr(PG_FUNCTION_ARGS);
/** @}*/

/** @{*/
//! Вывод таблицы проверки мандатной целостности кластера
/**
 * @return SETOF TEXT
 */
Datum pgmac_check_mac_integrity(PG_FUNCTION_ARGS);
/** @}*/

/** @{*/
//! Исправление мандатных атрибутов кластера
/**
 * @return void
 */
Datum pgmac_fix_mac_integrity(PG_FUNCTION_ARGS);
/** @}*/

/** @{*/
//! Аккумулирует мандатные метки.
/**
 * @return указатель на мандатную метку (prfree)
 */
Datum pgmac_accumulate_maclabel(PG_FUNCTION_ARGS);
/** @}*/

/**
 * @name Функции вспомогательные функции
 *
 */

//! Установка метки на файл
void pgac_chmac(const char *filename, const pgmac_label_t *label);

void pgac_init(void);

//! Инициализация большого объекта
void
pgac_initLargeObjectLabel (LargeObjectDesc *obj_desc);

#endif /* PGAC_MAC_H */
