#ifndef __PGMAC_TYPES_H
#define __PGMAC_TYPES_H
/**
 * @file
 * @brief pgac_types Описание типа мандатной метки и вектора мандатных атрибутов
 * @ingroup iutils
 *
 * Описания и определения объектов этого модуля расположены в файлах <tt>pgac_types.[hc]</tt>.
 *
 * Все внешние функции этого модуля, предназначенные для работы с типом мандатной
 * метки и подключаемые в <tt>include/catalog/pg_proc.h</tt>, могут иметь тип
 * <tt>IMMUTABLE ('i')</tt>, так как не зависят от иной информации, помимо своих
 * собственных аргументов и не имеют дополнительных эффектов (для справки см.
 * описание команды <tt>CREATE FUNCTION</tt> в "Описании команд").
 */

#include "postgres.h"
#include <fmgr.h>
#include <parsec/cap.h>
#include <inttypes.h>

/**
 * @name Вектор мандатных атрибутов
 */
/** @{*/
//! Вектор мандатных атрибутов
typedef int64 pgmac_caps_t;

//! Проверка на право менять метки данных
#define pgmac_CAPABLE_CHMAC(x)			((x)& PARSEC_CAP_TO_MASK(PARSEC_CAP_CHMAC))
//! Проверка на право менять текущую метку сессии
#define pgmac_CAPABLE_SETMAC(x)			((x)& PARSEC_CAP_TO_MASK(PARSEC_CAP_SETMAC))
//! Проверка на право игнорировать мандатный контроль при чтении данных
#define pgmac_CAPABLE_READSEARCH(x)		((x)& PARSEC_CAP_TO_MASK(PARSEC_CAP_READSEARCH))
//! Проверка на право игнорирования мандатного контроля по уровню конфиденциальности
#define pgmac_CAPABLE_IGNMACLVL(x)		((x)& PARSEC_CAP_TO_MASK(PARSEC_CAP_IGNMACLVL))
//! Проверка на право игнорирования мандатного контроля по вектору категорий доступа
#define pgmac_CAPABLE_IGNMACCAT(x)		((x)& PARSEC_CAP_TO_MASK(PARSEC_CAP_IGNMACCAT))
//! В СУБД в настоящее время не используется
//#define pgmac_CAPABLE_PRIV_SOCK(x)	((x)&CAP_PRIV_SOCK)

//! Удобный макрос для установки прав пользователя на игнорирование контроля и по
//! уровню конфидециальности и по вектору категорий доступа
#define pgmac_CAP_IGN					(PARSEC_CAP_TO_MASK(PARSEC_CAP_IGNMACLVL)|PARSEC_CAP_TO_MASK(PARSEC_CAP_IGNMACCAT))
//! Соответствующая pgmac_CAP_IGN проверка
#define pgmac_CAPABLE_IGN(x)			(pgmac_CAPABLE_IGNMACLVL (x) && pgmac_CAPABLE_IGNMACCAT (x))

//! Сложить два вектора мандатных атрибутов
#define pgmac_CAPS_ADD(x,y)				((x) | (y))
//! Вычесть второй вектор мандатных атрибутов из первого (т.е. сбросить установленные
//! в нем атрибуты в первом векторе в нулевое значение)
#define pgmac_CAPS_SUB(x,y)				((x) & ~(y))
/** @}*/

/**
 * @name Мандатная метка
 *
 * При сравнении меток с помощью макросов следует учитывать, что они сравниваются по правилам
 * мандатной модели, в которой существуют следующие варианты отношения двух меток:
 *   -# метка @a m1 выше метки @a m2;
 *   -# метка @a m1 ниже метки @a m2;
 *   -# метка @a m1 равна метке @a m2;
 *   -# метка @a m1 несравнима с меткой @a m2;
 *   -# комбинации вариантов 1 - 3.
 *
 * Комбинируя макросы pgmac_LE, pgmac_GE и pgmac_EQ можно проверить все возможные варианты
 * отношений между метками.
 */
/** @{*/
//! Уровень конфиденциальности
typedef uint8  pgmac_lev_t;
//! Вектор категорий доступа
typedef uint64 pgmac_cat_t;

//! Мандатная метка
/**
 *
 * @warning Без подсказки компилятору об упаковке этой структуры на разных архитектурах
 *          из-за выравнивания полей возможен разный реальный размер метки, отличный от
 *          суммы размеров полей. Это приводит к расхождению информации в
 *          <tt>include/catalog/pg_type.h</tt> и реального размера метки, что приводит
 *          к неправильному копированию меток во внутренних процедурах PGSQL.
 *
 * @warning При изменении размера этой структуры необходимо скорректировать соответствующее
 *          значение в записи о типе maclabel в <tt>include/catalog/pg_type.h</tt>.
 */
typedef struct pgmac_label_t {
	pgmac_lev_t lev; //!< Уровень конфиденциальности
	pgmac_cat_t cat; //!< Вектор категорий доступа
} __attribute__((packed)) pgmac_label_t;
//! Уровень конфиденциальности мандатной метки
#define pgmac_LEV(_x) ((_x)->lev)
//! Вектор категорий доступа мандатной метки
#define pgmac_CAT(_x) ((_x)->cat)

//! Для использования в определении системных таблиц (в частности --- pg_class)
typedef pgmac_label_t maclabel;

//! Метка @a _l1 ниже или равна метке @a _l2
#define pgmac_LE(_l1, _l2)			((pgmac_LEV(_l1) <= pgmac_LEV(_l2)) && ((pgmac_CAT(_l1) & pgmac_CAT(_l2)) == pgmac_CAT(_l1)))
//! Метка @a _l1 выше или равна метке @a _l2
#define pgmac_GE(_l1, _l2)			((pgmac_LEV(_l1) >= pgmac_LEV(_l2)) && ((pgmac_CAT(_l1) & pgmac_CAT(_l2)) == pgmac_CAT(_l2)))
//! Метка @a _l1 равна @a метке _l2
#define pgmac_EQ(_l1, _l2)			((pgmac_LEV(_l1) == pgmac_LEV(_l2)) && (pgmac_CAT(_l1) == pgmac_CAT(_l2)))
//! Определение максимальной (наиболее полной метки из двух меток)
//! После выполнения в l1 будет содержаться наиболее полная метка
#define pgmac_MAX(_l1, _l2) \
do { \
	if (pgmac_LEV(_l1) < pgmac_LEV(_l2)) \
		pgmac_LEV(_l1) = pgmac_LEV(_l2); \
	pgmac_CAT(_l1) |= pgmac_CAT(_l2); \
} while (0);

//! Определение значения нулевой мандатной метки
#define pgmac_DECLARE_ZERO_LABEL {0,0x0ULL}
//! Определение текстового представления нулевой мандатной метки
#define ZERO_MACLABEL "{0,0}"
//! Константа для установки метки в нулевое значение
extern const pgmac_label_t pgmac_zero_label;
//! Нулевое значение метки
#define pgmac_ZERO_P() (&pgmac_zero_label)

//! Определение значения неопределенной мандатной метки
#define pgmac_DECLARE_UNDEFINED_LABEL {255,0xFFFFFFFFFFFFFFFFULL}
//! Константа для установки метки в неопределенное значение и соответствующей проверки
extern const pgmac_label_t pgmac_undefined_label;

//! Определение формата текстового представления мандатной метки
#define FMTMACLABEL		"{%hhu,%" PRIX64 "}"
#define MAXMACLABELLEN	32

//! Получение указателя на мандатную метку из Datum
#define pgmac_DatumGet_P(_x)		((pgmac_label_t *)DatumGetPointer (_x))
//! Преобразование указателя на мандатную метку в Datum
#define pgmac_GetDatum_P(_x)		(PointerGetDatum (_x))
//! Получить указатель на мандатную метку со стека в функции
#define pgmac_GETARG_P(_n)			pgmac_DatumGet_P (PG_GETARG_DATUM (_n))
//! Вернуть мандатную метку как результат работы функции
#define pgmac_RETURN_P(_x)			return pgmac_GetDatum_P (_x)

//! Преобразование метки в C-строку
/**
 * @param label указатель на мандатную метку
 * @param buf указатель на буфер (если NULL, выделяется новый)
 * @param len размер буфера (>= MAXMACLABELLEN)
 * @return указатель на C-строку (prfee если выделялось)
 */
char* pgmac_fmt(pgmac_label_t *label, char *buf, int len);

//! Преобразование C-строки в метку
/**
 * @param 0 указатель на C-строку
 * @return указатель на мандатную метку (после изпользования следует освободить
 *			занимаемую память с помощью pfree)
 */
Datum pgmac_in(PG_FUNCTION_ARGS);

//! Преобразование метки в C-строку
/**
 * @param 0 указатель на мандатную метку
 * @return указатель на C-строку (prfee)
 */
Datum pgmac_out(PG_FUNCTION_ARGS);

//! Загрузка мандатной метки из бинарного потока
/**
 * @param 0 указатель на поток байт
 * @return указатель на мандатную метку (после изпользования следует освободить
 *         занимаемую память с помощью pfree)
 */
Datum pgmac_recv(PG_FUNCTION_ARGS);

//! Вывод мандатной метки в бинарный поток
/**
 * @param 0 указатель на мандатную метку
 * @return указатель на бинарный буфер
 */
Datum pgmac_send(PG_FUNCTION_ARGS);

//! Преобразование типа TEXT в мандатную метку
/**
 * @param 0 указатель на text
 * @return указатель на мандатную метку (pfree)
 */
Datum pgmac_text2mac(PG_FUNCTION_ARGS);

//! Преобразование мандатной метки в тип TEXT
/**
 * @param 0 указатель на мандатную метку
 * @return указатель на text (pfree)
 */
Datum pgmac_mac2text(PG_FUNCTION_ARGS);

//! Получение уровня конфиденциальности
/**
 * @param 0 указатель на мандатную метку
 * @return 32-х разрядное целое (int4)
 */
Datum pgmac_level(PG_FUNCTION_ARGS);

//! Получение вектора категорий доступа
/**
 * @param 0 указатель на мандатную метку
 * @return 64-х разрядное целое (int8)
 */
Datum pgmac_category(PG_FUNCTION_ARGS);

//! Создание метки из уровня конфиденциальности и вектора категорий доступа,
//! заданных как целые числа
/**
 * @param 0 32-х разрядное значение уровня конфиденциальности (int4)
 * @param 1 64-х разрядное значение вектора категорий (int8)
 * @return указатель на мандатную метку (prfee)
 */
Datum pgmac_maclabel(PG_FUNCTION_ARGS);

//! Равенство меток
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если метка @a 0 равна метке @a 1
 */
Datum pgmac_eq(PG_FUNCTION_ARGS);

//! Метка @a 0 ниже метки @a 1
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 ниже @a 1
 */
Datum pgmac_lt(PG_FUNCTION_ARGS);

//! Метка @a 0 ниже или равна метке @a 1
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 ниже или равна @a 1
 */
Datum pgmac_le(PG_FUNCTION_ARGS);

//! Метка @a 0 выше или равна метке @a 1
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 выше или равна @a 1
 */
Datum pgmac_ge(PG_FUNCTION_ARGS);

//! Метка @a 0 выше метки @a 1
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 выше @a 1
 */
Datum pgmac_gt(PG_FUNCTION_ARGS);

//! Метка @a 0 не равна метке @a 1
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 не равна @a 1
 */

//! Негаторы для операторов <, <=, >, >=
Datum pgmac_not_lt(PG_FUNCTION_ARGS);
Datum pgmac_not_le(PG_FUNCTION_ARGS);
Datum pgmac_not_gt(PG_FUNCTION_ARGS);
Datum pgmac_not_ge(PG_FUNCTION_ARGS);
Datum pgmac_ne(PG_FUNCTION_ARGS);
//! Метки @a 0 и @a 1 несравнимы
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 несравнима с @a 1
 */
Datum pgmac_nc(PG_FUNCTION_ARGS);
/** @}*/

/**
 * @name Функции поддержки индексирования полей типа мандатная метка
 *
 * Эти процедуры сравнения служат только для однозначного упорядочивания меток
 * и никак не связаны с логикой мандатного контроля.
 *
 * Во всех этих процедурах метка рассматривается как большое целое число, состоящее из
 * двух частей.
 *
 * @note Несмотря на то, что процедуры равенства и неравенства для индексов по реализации
 *       соответствуют мандатной модели, тем не менее, они выделены, чтобы обеспечить полную
 *       развязку операторов для работы индексов и операций логического сравнения меток.
 */
/** @{*/
//! Метка @a 0 равна метке @a 1
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 равна @a 1
 */
Datum pgmac_btree_eq(PG_FUNCTION_ARGS);

//! Метка @a 0 меньше метки @a 1
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 меньше @a 1
 */
Datum pgmac_btree_lt(PG_FUNCTION_ARGS);

//! Метка @a 0 меньше или равна метке @a 1
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 меньше или равна @a 1
 */
Datum pgmac_btree_idx(PG_FUNCTION_ARGS);

//! Метка @a 0 меньше или равна метке @a 1
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 меньше или равна @a 1
 */
Datum pgmac_btree_le(PG_FUNCTION_ARGS);

//! Метка @a 0 больше или равна метке @a 1
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 больше или равна @a 1
 */
Datum pgmac_btree_ge(PG_FUNCTION_ARGS);

//! Метка @a 0 больше метки @a 1
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 больше @a 1
 */
Datum pgmac_btree_gt(PG_FUNCTION_ARGS);

//! Метка @a 0 не равна метке @a 1
/**
 * @param 0 метка @a 0
 * @param 1 метка @a 1
 * @return @c true если @a 0 не равна @a 1
 */
Datum pgmac_btree_ne(PG_FUNCTION_ARGS);

//! Универсальная функция сравнения меток
/**
 * @param 0 указатель на мандатную метку
 * @param 1 указатель на мандатную метку
 * @return целое 32-х разрядное число:
 *         - -1, если метка @a 0 меньше метки @a 1
 *         -  0, если метки равны
 *         - -1, если метка @a 0 больше метки @a 1
 */
Datum pgmac_btree_cmp(PG_FUNCTION_ARGS);

//! Вычисление хэш-числа для метки, которое необходимо для построения хэш-индексов
/**
 * @param 0 метка
 * @return хэш-число
 */
Datum pgmac_btree_hash(PG_FUNCTION_ARGS);
/** @}*/

#endif /* __PGMAC_TYPES_H */
